searchNodes=[{"ref":"AyeSQL.html","title":"AyeSQL","type":"module","doc":"Aye /ʌɪ/ exclamation (archaic dialect): said to express assent; yes. AyeSQL is a library for using raw SQL. Overview Inspired by Clojure library Yesql, AyeSQL tries to find a middle ground between strings with raw SQL queries and SQL DSLs by: Keeping SQL in SQL files. Generating Elixir functions for every query. Supporting mandatory and optional named parameters. Allowing query composability with ease. Working out of the box with PostgreSQL using Ecto or Postgrex: Being extended to support other databases via the behaviour AyeSQL.Runner. Small Example Let&#39;s say we have an SQL query to retrieve the click count of a certain type of link every day of the last X days. In raw SQL this could be written as: WITH computed_dates AS ( SELECT dates::date AS date FROM generate_series( current_date - $1::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ) AS dates ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_dates AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = $2 GROUP BY dates.date ORDER BY dates.date; The equivalent query in Ecto would be: dates = ~s( SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS d ) from( c in &quot;clicks&quot;, right_join: day in fragment(dates, ^days), on: day.d == fragment(&quot;date(?)&quot;, c.inserted_at), where: c.link_id = ^link_id group_by: day.d, order_by: day.d, select: %{ day: fragment(&quot;date(?)&quot;, day.d), count: count(c.id) } ) Using fragments can get convoluted and difficult to maintain. In AyeSQL, the equivalent would be to create an SQL file with the query e.g. queries.sql: -- name: get_day_interval SELECT datetime::date AS date FROM generate_series( current_date - :days::interval, -- Named parameter :days current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ); -- name: get_avg_clicks -- docs: Gets average click count. WITH computed_dates AS ( :get_day_interval ) -- Composing with another query SELECT dates.date AS day, count(clicks.id) AS count FROM computed_date AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = :link_id -- Named parameter :link_id GROUP BY dates.date ORDER BY dates.date; In Elixir, we would load all the queries in this file by creating the following module: defmodule Queries do use AyeSQL, repo: MyRepo defqueries(&quot;queries.sql&quot;) # File name with relative path to SQL file. end or using the macro defqueries/3: import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;queries.sql&quot;, repo: MyRepo) Both approaches will create a module called Queries with all the queries defined in queries.sql. And then we could execute the query as follows: iex&gt; params = [ ...&gt; link_id: 42, ...&gt; days: %Postgrex.Interval{secs: 864_000} # 10 days ...&gt; ] iex&gt; Queries.get_avg_clicks(params, run?: true) {:ok, [ %{day: ..., count: ...}, %{day: ..., count: ...}, %{day: ..., count: ...}, ... ] }"},{"ref":"AyeSQL.html#__using__/1","title":"AyeSQL.__using__/1","type":"macro","doc":"Uses AyeSQL for loading queries. By default, supports the option runner (see AyeSQL.Runner behaviour). Any other option will be passed to the runner."},{"ref":"AyeSQL.html#defqueries/1","title":"AyeSQL.defqueries/1","type":"macro","doc":"Macro to load queries from a file. Let&#39;s say we have the file lib/sql/queries.sql with the following contents: -- name: get_user -- docs: Gets user by username SELECT * FROM users WHERE username = :username; Then we can load our queries to Elixir using the macro defqueries/1: # file: lib/queries.ex defmodule Queries do use AyeSQL, repo: MyRepo defqueries(&quot;sql/queries.sql&quot;) end or the macro defqueries/3: # file: lib/queries.ex import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;sql/queries.ex&quot;, repo: MyRepo) And finally we can inspect the query: iex(1)&gt; Queries.get_user!(username: &quot;some_user&quot;) {:ok, %AyeSQL.Query{ statement: &quot;SELECT * FROM user WHERE username = $1&quot;, arguments: [&quot;some_user&quot;] } } or run it: iex(1)&gt; Queries.get_user!(username: &quot;some_user&quot;, run?: true) {:ok, [ %{username: ..., ...} ] } For running it by default, we can set the following in our configuration: config :ayesql, run?: true"},{"ref":"AyeSQL.html#defqueries/3","title":"AyeSQL.defqueries/3","type":"macro","doc":"Macro to load queries from a file and create a module for them. Same as defqueries/1, but creates a module e.g for the query file lib/sql/queries.sql we can use this macro as follows: # file: lib/queries.ex import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;sql/queries.sql&quot;, repo: MyRepo) This will generate the module Queries and it&#39;ll contain all the SQL statements included in sql/queries.sql."},{"ref":"AyeSQL.Core.html","title":"AyeSQL.Core","type":"module","doc":"This module defines the core functionality for AyeSQL."},{"ref":"AyeSQL.Core.html#run?/0","title":"AyeSQL.Core.run?/0","type":"function","doc":"Whether the queries should run by default or not. You can set this in the configuration as: use Mix.Config config :ayesql, run?: true"},{"ref":"AyeSQL.Core.html#t:options/0","title":"AyeSQL.Core.options/0","type":"type","doc":"AyeSQL query options."},{"ref":"AyeSQL.Core.html#t:parameter_name/0","title":"AyeSQL.Core.parameter_name/0","type":"type","doc":"AyeSQL query parameter name."},{"ref":"AyeSQL.Core.html#t:parameters/0","title":"AyeSQL.Core.parameters/0","type":"type","doc":"AyeSQL query parameters."},{"ref":"AyeSQL.Error.html","title":"AyeSQL.Error","type":"module","doc":"This module defines an AyeSQL error."},{"ref":"AyeSQL.Error.html#__struct__/0","title":"AyeSQL.Error.__struct__/0","type":"function","doc":"A query struct."},{"ref":"AyeSQL.Error.html#new/1","title":"AyeSQL.Error.new/1","type":"function","doc":"Creates a new error given some options."},{"ref":"AyeSQL.Error.html#t:arguments/0","title":"AyeSQL.Error.arguments/0","type":"type","doc":"Query arguments."},{"ref":"AyeSQL.Error.html#t:errors/0","title":"AyeSQL.Error.errors/0","type":"type","doc":"Query errors."},{"ref":"AyeSQL.Error.html#t:statement/0","title":"AyeSQL.Error.statement/0","type":"type","doc":"Query statement."},{"ref":"AyeSQL.Error.html#t:t/0","title":"AyeSQL.Error.t/0","type":"type","doc":"An error type."},{"ref":"AyeSQL.Query.html","title":"AyeSQL.Query","type":"module","doc":"This module defines an AyeSQL query."},{"ref":"AyeSQL.Query.html#__struct__/0","title":"AyeSQL.Query.__struct__/0","type":"function","doc":"A query struct."},{"ref":"AyeSQL.Query.html#new/1","title":"AyeSQL.Query.new/1","type":"function","doc":"Creates a new query given some options."},{"ref":"AyeSQL.Query.html#t:arguments/0","title":"AyeSQL.Query.arguments/0","type":"type","doc":"Query arguments."},{"ref":"AyeSQL.Query.html#t:statement/0","title":"AyeSQL.Query.statement/0","type":"type","doc":"Query statement."},{"ref":"AyeSQL.Query.html#t:t/0","title":"AyeSQL.Query.t/0","type":"type","doc":"A query type."},{"ref":"AyeSQL.AST.html","title":"AyeSQL.AST","type":"module","doc":"This module defines function for expanding the AST."},{"ref":"AyeSQL.AST.html#t:expand_function/0","title":"AyeSQL.AST.expand_function/0","type":"type","doc":"Function to be applied with some parameters and context."},{"ref":"AyeSQL.AST.Context.html","title":"AyeSQL.AST.Context","type":"module","doc":"This module defines an AST context."},{"ref":"AyeSQL.AST.Context.html#__struct__/0","title":"AyeSQL.AST.Context.__struct__/0","type":"function","doc":"AST context struct."},{"ref":"AyeSQL.AST.Context.html#add_index/2","title":"AyeSQL.AST.Context.add_index/2","type":"function","doc":"Adds a value to the context index."},{"ref":"AyeSQL.AST.Context.html#id/1","title":"AyeSQL.AST.Context.id/1","type":"function","doc":"Context id function."},{"ref":"AyeSQL.AST.Context.html#merge/2","title":"AyeSQL.AST.Context.merge/2","type":"function","doc":"Merges two contexts."},{"ref":"AyeSQL.AST.Context.html#merge_error/2","title":"AyeSQL.AST.Context.merge_error/2","type":"function","doc":"Merges a context with an error"},{"ref":"AyeSQL.AST.Context.html#merge_query/2","title":"AyeSQL.AST.Context.merge_query/2","type":"function","doc":"Merges a context with a query."},{"ref":"AyeSQL.AST.Context.html#new/1","title":"AyeSQL.AST.Context.new/1","type":"function","doc":"Creates a new context given some options."},{"ref":"AyeSQL.AST.Context.html#not_found/2","title":"AyeSQL.AST.Context.not_found/2","type":"function","doc":"Updates context with the error not found for a key."},{"ref":"AyeSQL.AST.Context.html#put_argument/2","title":"AyeSQL.AST.Context.put_argument/2","type":"function","doc":"Adds arguments in a context given a new value."},{"ref":"AyeSQL.AST.Context.html#put_statement/2","title":"AyeSQL.AST.Context.put_statement/2","type":"function","doc":"Adds statement in a context given a new value."},{"ref":"AyeSQL.AST.Context.html#put_variable/2","title":"AyeSQL.AST.Context.put_variable/2","type":"function","doc":"Puts a new variable value in the context."},{"ref":"AyeSQL.AST.Context.html#put_variables/2","title":"AyeSQL.AST.Context.put_variables/2","type":"function","doc":"Puts several variable value in the context as an SQL list."},{"ref":"AyeSQL.AST.Context.html#to_query/1","title":"AyeSQL.AST.Context.to_query/1","type":"function","doc":"Transforms a context to a query."},{"ref":"AyeSQL.AST.Context.html#t:arguments/0","title":"AyeSQL.AST.Context.arguments/0","type":"type","doc":"Argument list."},{"ref":"AyeSQL.AST.Context.html#t:error/0","title":"AyeSQL.AST.Context.error/0","type":"type","doc":"Error."},{"ref":"AyeSQL.AST.Context.html#t:error_type/0","title":"AyeSQL.AST.Context.error_type/0","type":"type","doc":"Error type."},{"ref":"AyeSQL.AST.Context.html#t:index/0","title":"AyeSQL.AST.Context.index/0","type":"type","doc":"Current context index."},{"ref":"AyeSQL.AST.Context.html#t:statement/0","title":"AyeSQL.AST.Context.statement/0","type":"type","doc":"Accumulated statement."},{"ref":"AyeSQL.AST.Context.html#t:t/0","title":"AyeSQL.AST.Context.t/0","type":"type","doc":"AST context."},{"ref":"AyeSQL.Parser.html","title":"AyeSQL.Parser","type":"module","doc":"This module defines AyeSQL file parsing."},{"ref":"AyeSQL.Parser.html#create_queries/1","title":"AyeSQL.Parser.create_queries/1","type":"function","doc":"Creates several queries from the contents of a file."},{"ref":"AyeSQL.Parser.html#t:content/0","title":"AyeSQL.Parser.content/0","type":"type","doc":"Contents of the query."},{"ref":"AyeSQL.Parser.html#t:docs/0","title":"AyeSQL.Parser.docs/0","type":"type","doc":"Documentation of the query function."},{"ref":"AyeSQL.Parser.html#t:filename/0","title":"AyeSQL.Parser.filename/0","type":"type","doc":"File name."},{"ref":"AyeSQL.Parser.html#t:fun_def/0","title":"AyeSQL.Parser.fun_def/0","type":"type","doc":"Definition of the query in the AST."},{"ref":"AyeSQL.Parser.html#t:name/0","title":"AyeSQL.Parser.name/0","type":"type","doc":"Name of the query function."},{"ref":"AyeSQL.Runner.html","title":"AyeSQL.Runner","type":"behaviour","doc":"This module defines an AyeSQL.Runner."},{"ref":"AyeSQL.Runner.html#__using__/1","title":"AyeSQL.Runner.__using__/1","type":"macro","doc":"Uses the AyeSQL.Runner behaviour."},{"ref":"AyeSQL.Runner.html#c:run/2","title":"AyeSQL.Runner.run/2","type":"callback","doc":"Callback to initialize the runner."},{"ref":"AyeSQL.Runner.Ecto.html","title":"AyeSQL.Runner.Ecto","type":"module","doc":"This module defines Ecto default adapter. Can be used as follows: defmodule MyQueries do use AyeSQL, repo: MyRepo defqueries(&quot;query/my_queries.sql&quot;) end"},{"ref":"AyeSQL.Runner.Postgrex.html","title":"AyeSQL.Runner.Postgrex","type":"module","doc":"This module defines Postgrex default adapter. Can be used as follows: defmodule MyQueries do use AyeSQL, runner: AyeSQL.Runner.Postgrex defqueries(&quot;query/my_queries.sql&quot;) end And given a connection to the database, then it can be used with the query options: iex&gt; MyQueries.get_user([id: id], run?: true, conn: connection) {:ok, ...}"},{"ref":"readme.html","title":"AyeSQL","type":"extras","doc":"AyeSQL Aye /ʌɪ/ exclamation (archaic dialect): said to express assent; yes. AyeSQL is a library for using raw SQL."},{"ref":"readme.html#overview","title":"AyeSQL - Overview","type":"extras","doc":"Inspired by Clojure library Yesql, AyeSQL tries to find a middle ground between strings with raw SQL queries and SQL DSLs by: Keeping SQL in SQL files. Generating Elixir functions for every query. Supporting mandatory and optional named parameters. Allowing query composability with ease. Working out of the box with PostgreSQL using Ecto or Postgrex. Being extended to support other databases via the behaviour AyeSQL.Runner. If you want to know more why this project exists: SQL in Elixir Why raw SQL? AyeSQL: Writing Raw SQL in Elixir (External Link) If you want to know more about AyeSQL: Small example Syntax Naming queries Named parameters Mandatory parameters Query composition Optional parameters IN statement Subqueries Query runners Running queries by default Installation"},{"ref":"readme.html#sql-in-elixir","title":"AyeSQL - SQL in Elixir","type":"extras","doc":"Writing and running raw SQL in Elixir is not pretty. Not only the lack of syntax highlighting is horrible, but also substituting parameters into the query string can be unmaintainable e.g: query = &quot;&quot;&quot; SELECT hostname, AVG(ram_usage) AS avg_ram FROM server WHERE hostname IN ($1, $2, $3) AND location = $4 GROUP BY hostname &quot;&quot;&quot; arguments = [&quot;server_0&quot;, &quot;server_1&quot;, &quot;server_2&quot;, &quot;Barcelona&quot;] Postgrex.query(conn, query, arguments) Adding more hostnames to the previous query would be a nightmare. If the arguments are generated dynamically, then editing this query would be a challenging task. Thankfully, we have Ecto, which provides a great DSL for generating database queries at runtime. The same query in Ecto would be the following: servers = [&quot;server_0&quot;, &quot;server_1&quot;, &quot;server_2&quot;] location = &quot;Barcelona&quot; from s in &quot;server&quot;, where: s.location == ^location and s.hostname in ^servers, select: %{hostname: s.hostname, avg_ram: avg(s.ram_usage)} Pretty straightforward and maintainable."},{"ref":"readme.html#why-raw-sql","title":"AyeSQL - Why raw SQL","type":"extras","doc":"If Ecto is so good for building queries, why would you use raw SQL?. Though Ecto is quite good with simple queries, complex custom queries often require the use of fragments. Fragments are not pretty though there are workarounds using macros to make them prettier. It&#39;s easier to see with an example: let&#39;s say we have to [retrieve the click count of a certain type of link every day of the last N days]((https://stackoverflow.com/questions/39556763/use-ecto-to-generate-series-in-postgres-and-also-retrieve-null-values-as-0). With complex queries, developers tend to start writing them in raw SQL: WITH computed_dates AS ( SELECT dates::date AS date FROM generate_series( current_date - $1::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ) AS dates ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_dates AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = $2 GROUP BY dates.date ORDER BY dates.date; Once we have the raw SQL, it&#39;s a bit easier to write our Ecto query. In this case, this query should be written using fragments: dates = &quot;&quot;&quot; SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS d &quot;&quot;&quot; from( c in &quot;clicks&quot;, right_join: day in fragment(dates, ^days), on: day.d == fragment(&quot;date(?)&quot;, c.inserted_at), where: c.link_id = ^link_id group_by: day.d, order_by: day.d, select: %{ day: fragment(&quot;date(?)&quot;, day.d), count: count(c.id) } ) It&#39;s not the ideal solution (yet) and it&#39;s harder to maintain than the raw SQL solution that would work out-of-the-box. Only to get to the previous solution, the developer need to: Know the specific SQL dialect of the database they&#39;re using. Know Ecto&#39;s API and its limitations. For both, raw SQL and Ecto query, the end result for this query would be the same. With extra effort we found a subpar solution that gives us the same result as our raw SQL. The final, and sometimes, optional step would be to transform the Ecto query into something a bit more maintainable. defmodule CustomDSL do defmacro date(date) do quote do fragment(&quot;date(?)&quot;, unquote(date)) end end defmacro ndays(n) do query = &quot;&quot;&quot; SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS d &quot;&quot;&quot; quote do fragment(unquote(query), unquote(n)) end end end import CustomDSL from( c in &quot;clicks&quot;, right_join: day in ndays(^days), on: day.d == date(c.inserted_at), where: c.link_id = ^link_id group_by: day.d, order_by: day.d, select: %{ day: date(day.d) count: count(c.id) } ) The previous query is more readable, but requires knowledge of: The specific SQL dialect. Ecto&#39;s API and its limitations. Elixir&#39;s macros. Custom DSL API. For some problems, getting to this final stage is preferrable. However, for some other problems, the raw SQL query would have been enough. The raw SQL query was already a good solution to the problem. It only needs a maintainable way to be parametrized. That&#39;s why AyeSQL exists."},{"ref":"readme.html#small-example","title":"AyeSQL - Small Example","type":"extras","doc":"In AyeSQL, the equivalent would be to create an SQL file with the query e.g. queries.sql: -- file: queries.sql -- name: get_avg_clicks -- docs: Gets average click count. WITH computed_dates AS ( SELECT datetime::date AS date FROM generate_series( current_date - :days::interval, -- Named parameter :days current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ) ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_date AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = :link_id -- Named parameter :link_id GROUP BY dates.date ORDER BY dates.date; In Elixir, we would load all the queries in this file by creating the following module: # file: lib/queries.ex defmodule Queries do use AyeSQL, repo: MyRepo defqueries(&quot;queries.sql&quot;) # File name with relative path to SQL file. end or using the macro defqueries/3: # file: lib/queries.ex import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;queries.sql&quot;, repo: MyRepo) Note: The file name used in defqueries macro should be relative to the file where the macro is used. Both approaches will create a module called Queries with all the queries defined in queries.sql. And then we could execute the query as follows: iex&gt; params = [ ...&gt; link_id: 42, ...&gt; days: %Postgrex.Interval{secs: 864_000} # 10 days ...&gt; ] iex&gt; Queries.get_avg_clicks(params, run?: true) {:ok, [ %{day: ..., count: ...}, %{day: ..., count: ...}, %{day: ..., count: ...}, ... ] }"},{"ref":"readme.html#syntax","title":"AyeSQL - Syntax","type":"extras","doc":"An SQL file can have as many queries as you want as long as they are named. For the following sections we&#39;ll assume we have: lib/my_repo.ex which is an Ecto repo called MyRepo. lib/queries.sql with SQL queries. lib/queries.ex with the following structure: import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;queries.sql&quot;, repo: MyRepo) Running the queries by default by adding the following in config/config.exs: config ayesql, run?: true Naming Queries For naming queries, we add a comment with the keyword -- name: followed by the name of the function e.g the following query would generate the function Queries.get_hostnames/2: -- name: get_hostnames SELECT hostname FROM server Additionally, we could also add documentation for the query by adding a comment with the keyword -- docs: followed by the query&#39;s documentation e.g: -- name: get_hostnames -- docs: Gets hostnames from the servers. SELECT hostname FROM server Important: if the function does not have -- docs: it won&#39;t have documentation e.g. @doc false. Named Parameters There are two types of named parameters: Mandatory: for passing parameters to a query. They start with : e.g. :hostname. Optional: for query composability. The start with :_ e.g. :_order_by. Additionaly, any query in a file can be accessed with its name adding : at the front e.g :get_hostnames. Mandatory Parameters Let&#39;s say we want to get the name of an operative system by architecture: -- name: get_os_by_architecture -- docs: Gets operative system&#39;s name by a given architecture. SELECT name FROM operative_system WHERE architecture = :architecture The previous query would generate the function Queries.get_os_by_architecture/2 that can be called as: iex&gt; Queries.get_os_by_architecture(architecture: &quot;AMD64&quot;) {:ok, [ %{name: &quot;Debian Buster&quot;}, %{name: &quot;Windows 10&quot;}, ... ] } Query Composition Now if we would like to get hostnames by architecture we could compose queries by doing the following: -- name: get_os_by_architecture -- docs: Gets operative system&#39;s name by a given architecture. SELECT name FROM operative_system WHERE architecture = :architecture -- name: get_hostnames_by_architecture -- docs: Gets hostnames by architecture. SELECT hostname FROM servers WHERE os_name IN ( :get_os_by_architecture ) The previous query would generate the function Queries.get_hostnames_by_architecture/2 that can be called as: iex&gt; Queries.get_hostnames_by_architecture(architecture: &quot;AMD64&quot;) {:ok, [ %{hostname: &quot;server0&quot;}, %{hostname: &quot;server1&quot;}, ... ] } Optional Parameters Let&#39;s say that now we need to order ascending or descending by hostname by using an optional :_order_by parameter e.g: -- name: get_os_by_architecture -- docs: Gets operative system&#39;s name by a given architecture. SELECT name FROM operative_system WHERE architecture = :architecture -- name: get_hostnames_by_architecture -- docs: Gets hostnames by architecture. SELECT hostname FROM servers WHERE os_name IN ( :get_os_by_architecture ) :_order_by -- name: ascending ORDER BY hostname ASC -- name: descending ORDER BY hostname DESC The previous query could be called as before: iex&gt; Queries.get_hostnames_by_architecture(architecture: &quot;AMD64&quot;) {:ok, [ %{hostname: &quot;Barcelona&quot;}, %{hostname: &quot;Granada&quot;}, %{hostname: &quot;Madrid&quot;}, ... ] } or by order ascending: iex&gt; params = [architecture: &quot;AMD64&quot;, _order_by: :ascending] iex&gt; Queries.get_hostnames_by_architecture(params) {:ok, [ %{hostname: &quot;Barcelona&quot;}, %{hostname: &quot;Madrid&quot;}, %{hostname: &quot;Granada&quot;}, ... ] } or descending: iex&gt; params = [architecture: &quot;AMD64&quot;, _order_by: :descending] iex&gt; Queries.get_hostnames_by_architecture(params) {:ok, [ %{hostname: &quot;Zaragoza&quot;}, %{hostname: &quot;Madrid&quot;}, %{hostname: &quot;Granada&quot;}, ... ] } Important: A query can be called by name e.g. :descending if it&#39;s defined in the same SQL file. Otherwise, we need to pass the function instead e.g. Queries.descending/2 iex&gt; params = [architecture: &quot;AMD64&quot;, _order_by: Queries.descending/2] iex&gt; Queries.get_hostnames_by_architecture(params) {:ok, [ %{hostname: &quot;Zaragoza&quot;}, %{hostname: &quot;Madrid&quot;}, %{hostname: &quot;Granada&quot;}, ... ] } IN Statement Lists in SQL might be tricky. That&#39;s why AyeSQL supports a special type for them e.g: Let&#39;s say we have the following query: -- name: get_os_by_hostname -- docs: Gets hostnames and OS names given a list of hostnames. SELECT hostname, os_name FROM servers WHERE hostname IN (:hostnames) It is possible to do the following: iex&gt; params = [hostnames: {:in, [&quot;server0&quot;, &quot;server1&quot;, &quot;server2&quot;]}] iex&gt; Server.get_os_by_hostname(params) {:ok, [ %{hostname: &quot;server0&quot;, os_name: &quot;Debian Buster&quot;}, %{hostname: &quot;server1&quot;, avg_ram: &quot;Windows 10&quot;}, %{hostname: &quot;server2&quot;, avg_ram: &quot;Minix 3&quot;} ] } Subqueries Subqueries can be composed directly, as show before, or via the :inner tuple e.g. let&#39;s say we need to get the adults order by name in ascending order and age in descending order: -- name: ascending ASC -- name: descending DESC -- name: by_age age :order_direction -- name: by_name name :order_direction -- name: get_adults -- docs: Gets adults. SELECT name, age FROM person WHERE age &gt;= 18 ORDER BY :order_by Then our code in elixir would be: iex&gt; order_by = [ ...&gt; by_name: [order_direction: :ascending], ...&gt; by_age: [order_direction: :descending] ...&gt; ] iex&gt; Queries.get_adults(order_by: {:inner, order_by, &quot;, &quot;}) {:ok, [ %{name: &quot;Alice&quot;, age: 42}, %{name: &quot;Bob&quot;, age: 21}, ... ] } Note: If you&#39;re using this level of composability, consider using Ecto if it fits your problem or stack."},{"ref":"readme.html#query-runners","title":"AyeSQL - Query Runners","type":"extras","doc":"The purpose of runners is to be able to implement other database adapters. By default, AyeSQL uses the runner AyeSQL.Runner.Ecto. This runner only has one option which is :repo for the repo module. Additionally, it converts the result to a list of maps. Using other runners is as easy as setting them in the module definition as follows: defmodule Queries do use AyeSQL, runner: IdemRunner, repository: MyRepo defqueries(&quot;queries.sql&quot;) end or import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;queries.sql&quot;, runner: IdemRunner, repository: MyRepo) For runners, there is only one callback to be implemented. run/2: which receives a AyeSQL.Query.t() and a keyword() list with extra options for the runner. The following would be a runner for Ecto that does nothing to the result (returns Postgrex.Result.t() and Postgrex.Error.t() structs): defmodule IdemRunner do use AyeSQL.Runner alias AyeSQL.Query @impl true def run(%Query{statement: stmt, arguments: args}, options) do repo = options[:repository] || raise ArgumentError, &quot;No repo defined&quot; Ecto.Adapters.SQL.query(repo, stmt, args) end end"},{"ref":"readme.html#running-queries-by-default","title":"AyeSQL - Running Queries by Default","type":"extras","doc":"Queries are not run by default, but the AyeSQL.Query.t() struct is returned instead. For running queries by default, we can add the following to the config: import Config config :ayesql, run?: true And then we don&#39;t need to specify the [run?: true] options for every query."},{"ref":"readme.html#installation","title":"AyeSQL - Installation","type":"extras","doc":"AyeSQL is available as a Hex package. To install, add it to your dependencies in your mix.exs file: def deps do [{:ayesql, &quot;~&gt; 0.6&quot;}] end If you&#39;re going to use any of the provided query runners, then you should add their dependencies as well: Add :ecto_sql for AyeSQL.Runner.Ecto (default runner). Add :postgrex for AyeSQL.Runner.Postgrex. Add :ecto_sql and :postgrex for running queries using Ecto in a PostgreSQL database."},{"ref":"readme.html#author","title":"AyeSQL - Author","type":"extras","doc":"Alexander de Sousa."},{"ref":"readme.html#license","title":"AyeSQL - License","type":"extras","doc":"AyeSQL is released under the MIT License. See the LICENSE file for further details."}]